/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "DebuggerCoreLinux.h"
#include "State.h"
#include "DebugEvent.h"

#include <boost/bind.hpp>

#include <QDebug>

#include <cerrno>
#include <cstring>

#include <asm/ldt.h>
#include <signal.h>
#include <sys/mman.h>
#include <sys/ptrace.h>
#include <sys/user.h>
#include <sys/wait.h>
#include <unistd.h>

// doesn't always seem to be defined in the headers
#ifndef PTRACE_GETSIGINFO
#define PTRACE_GETSIGINFO		static_cast<__ptrace_request>(0x4202)
#endif

#ifndef PTRACE_GET_THREAD_AREA
#define PTRACE_GET_THREAD_AREA	static_cast<__ptrace_request>(25)
#endif

#ifndef PTRACE_SET_THREAD_AREA
#define PTRACE_SET_THREAD_AREA	static_cast<__ptrace_request>(26)
#endif

//#define DEBUG_THREADS

//------------------------------------------------------------------------------
// Name: DebuggerCore()
// Desc: constructor
//------------------------------------------------------------------------------
DebuggerCore::DebuggerCore() {
#if defined(_SC_PAGESIZE)
	m_PageSize = sysconf(_SC_PAGESIZE);
#elif defined(_SC_PAGE_SIZE)
	m_PageSize = sysconf(_SC_PAGE_SIZE);
#else
	m_PageSize = PAGE_SIZE;
#endif
}

//------------------------------------------------------------------------------
// Name: pageSize() const
// Desc: returns the size of a page on this system
//------------------------------------------------------------------------------
edb::address_t DebuggerCore::pageSize() const {
	return m_PageSize;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
DebuggerCore::~DebuggerCore() {
	detach();
}

//------------------------------------------------------------------------------
// Name: waitForDebugEvent(DebugEvent &event, int msecs)
// Desc: waits for a debug event, msecs is a timeout
//      it will return false if an error or timeout occurs
//------------------------------------------------------------------------------
bool DebuggerCore::waitForDebugEvent(DebugEvent &event, int msecs) {
	if(attached()) {		
		int status;
		bool timeout;
		
		const edb::tid_t tid = waitpid_timeout(-1, &status, __WALL, msecs, timeout);
		if(!timeout) {
			if(tid > 0) {
			
				// avoid spurious messages from other PIDs we used to be tracing
				if(!m_Threads.contains(tid)) {
					return false;
				}
			
#ifdef DEBUG_THREADS
				if(WIFSTOPPED(status) && WSTOPSIG(status) == SIGTRAP) {
					pid_t new_tid;
					if(((status >> 16) & 0xffff) == PTRACE_EVENT_CLONE) {
						if(ptrace(PTRACE_GETEVENTMSG, tid, 0, &new_tid) != -1) {
							qDebug() << "[DebuggerCore] tracing new thread:" << new_tid;
							m_Threads.insert(new_tid, ThreadInfo());
							ptrace(PTRACE_CONT, new_tid, 0, 0);
						}
						ptrace(PTRACE_CONT, tid, 0, 0);
						return false;
					}
				}
#endif
				event = DebugEvent(status, pid(), tid);
								
				m_ActiveThread			= event.thread();
				m_Threads[tid].status	= status;
				return true;
			}
		}
	}
	return false;
}

//------------------------------------------------------------------------------
// Name: readData(edb::address_t address, bool &ok)
// Desc: 
// Note: this will fail on newer versions of linux if called from a 
//       different thread than the one which attached to process
//------------------------------------------------------------------------------
long DebuggerCore::readData(edb::address_t address, bool &ok) {
	errno = 0;
	const long v = ptrace(PTRACE_PEEKTEXT, pid(), address, 0);
	SET_OK(ok, v);
	return v;
}

//------------------------------------------------------------------------------
// Name: writeData(edb::address_t address, long value)
// Desc: 
//------------------------------------------------------------------------------
bool DebuggerCore::writeData(edb::address_t address, long value) {
	return ptrace(PTRACE_POKETEXT, pid(), address, value) != -1;
}

//------------------------------------------------------------------------------
// Name: attach(edb::pid_t pid)
// Desc: 
//------------------------------------------------------------------------------
bool DebuggerCore::attach(edb::pid_t pid) {
	detach();
	
	const long ret = ptrace(PTRACE_ATTACH, pid, 0, 0);
	if(ret == 0) {
		m_PID			= pid;
		m_ActiveThread	= pid;
		m_Threads.clear();
		m_Threads.insert(pid, ThreadInfo());
		
		// TODO: attach to all of the threads
	}
	
	return ret == 0;
}

//------------------------------------------------------------------------------
// Name: detach()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerCore::detach() {
	if(attached()) {
		clearBreakpoints();
		for(threadmap_t::const_iterator it = m_Threads.begin(); it != m_Threads.end(); ++it) {
			ptrace(PTRACE_DETACH, it.key(), 0, 0);
		}
		
		m_PID = 0;
		m_Threads.clear();
	}
}

//------------------------------------------------------------------------------
// Name: kill()
// Desc: 
//------------------------------------------------------------------------------
void DebuggerCore::kill() {
	if(attached()) {
		clearBreakpoints();
		ptrace(PTRACE_KILL, pid(), 0, 0);
		native_waitpid(pid(), 0, __WALL);
		m_PID = 0;
		m_Threads.clear();
	}
}

//------------------------------------------------------------------------------
// Name: pause()
// Desc: stops *all* threads of a process
//------------------------------------------------------------------------------
void DebuggerCore::pause() {
	if(attached()) {
		for(threadmap_t::const_iterator it = m_Threads.begin(); it != m_Threads.end(); ++it) {
			::kill(it.key(), SIGSTOP);
		}
	}
}

//------------------------------------------------------------------------------
// Name: resume(edb::EVENT_STATUS status)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerCore::resume(edb::EVENT_STATUS status) {
	// TOOD: assert that we are paused
	
	if(attached()) {
		if(status != edb::DEBUG_STOP) {
			int code = 0;
			const edb::tid_t tid = activeThread();
			if(status == edb::DEBUG_EXCEPTION_NOT_HANDLED) {
				const int s = m_Threads[tid].status;

				if(WIFSIGNALED(s)) {
					code = WTERMSIG(s);
				} else if(WIFSTOPPED(s)) {
					code = WSTOPSIG(s);
				}
			}
			ptrace(PTRACE_CONT, tid, 0, code);
		}
	}
}

//------------------------------------------------------------------------------
// Name: step(edb::EVENT_STATUS status)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerCore::step(edb::EVENT_STATUS status) {
	// TOOD: assert that we are paused
	
	if(attached()) {
		if(status != edb::DEBUG_STOP) {
			int code = 0;
			const edb::tid_t tid = activeThread();
			if(status == edb::DEBUG_EXCEPTION_NOT_HANDLED) {
				const int s = m_Threads[tid].status;

				if(WIFSIGNALED(s)) {
					code = WTERMSIG(s);
				} else if(WIFSTOPPED(s)) {
					code = WSTOPSIG(s);
				}
			}

			ptrace(PTRACE_SINGLESTEP, tid, 0, code);
		}
	}
}

//------------------------------------------------------------------------------
// Name: getState(State &state)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerCore::getState(State &state) {
	// TOOD: assert that we are paused
	
	if(attached()) {
		struct user_regs_struct regs;
		if(ptrace(PTRACE_GETREGS, activeThread(), 0, &regs) != -1) {
		#if defined(EDB_X86)		
			state.eax = regs.eax;
			state.ebx = regs.ebx;
			state.ecx = regs.ecx;
			state.edx = regs.edx;
			state.esp = regs.esp;
			state.ebp = regs.ebp;
			state.edi = regs.edi;
			state.esi = regs.esi;
			state.eip = regs.eip;
			state.eflags = regs.eflags;
			state.cs_ = regs.xcs;
			state.ds_ = regs.xds;
			state.es_ = regs.xes;
			state.fs_ = regs.xfs;
			state.gs_ = regs.xgs;
			state.ss_ = regs.xss;
			state.orig_eax = regs.orig_eax;
			
			struct user_desc desc;
			std::memset(&desc, 0, sizeof(desc));
			if(ptrace(PTRACE_GET_THREAD_AREA, activeThread(), (state.gs_ / LDT_ENTRY_SIZE), &desc) != -1) {
				state.gs_base = desc.base_addr;
			} else {
				state.gs_base = 0;
			}

			if(ptrace(PTRACE_GET_THREAD_AREA, activeThread(), (state.fs_ / LDT_ENTRY_SIZE), &desc) != -1) {
				state.fs_base = desc.base_addr;
			} else {
				state.fs_base = 0;
			}
		#elif defined(EDB_X86_64)
			state.rax = regs.rax;
			state.rbx = regs.rbx;
			state.rcx = regs.rcx;
			state.rdx = regs.rdx;
			state.rsp = regs.rsp;
			state.rbp = regs.rbp;
			state.rdi = regs.rdi;
			state.rsi = regs.rsi;
			state.rip = regs.rip;
			state.r8 = regs.r8;
			state.r9 = regs.r9;
			state.r10 = regs.r10;
			state.r11 = regs.r11;
			state.r12 = regs.r12;
			state.r13 = regs.r13;
			state.r14 = regs.r14;
			state.r15 = regs.r15;
			state.rflags = regs.eflags;
			state.cs = regs.cs;
			state.ds = regs.ds;
			state.es = regs.es;
			state.fs = regs.fs;
			state.gs = regs.gs;
			state.ss = regs.ss;
			state.orig_rax = regs.orig_rax;
			state.fs_base = regs.fs_base;
			state.gs_base = regs.gs_base;
		#endif
		}
		
		// floating point registers
		struct user_fpregs_struct fpregs;
		if(ptrace(PTRACE_GETFPREGS, activeThread(), 0, &fpregs) != -1) {
			State::i387Register *const p = reinterpret_cast<State::i387Register *>(fpregs.st_space);
			qCopy(&p[0], &p[8], state.st);
		}
		
		// debug registers
		state.dr[0] = ptrace(PTRACE_PEEKUSER, activeThread(), offsetof(user, u_debugreg[0]), 0);
		state.dr[1] = ptrace(PTRACE_PEEKUSER, activeThread(), offsetof(user, u_debugreg[1]), 0);
		state.dr[2] = ptrace(PTRACE_PEEKUSER, activeThread(), offsetof(user, u_debugreg[2]), 0);
		state.dr[3] = ptrace(PTRACE_PEEKUSER, activeThread(), offsetof(user, u_debugreg[3]), 0);
		state.dr[4] = 0;
		state.dr[5] = 0;
		state.dr[6] = ptrace(PTRACE_PEEKUSER, activeThread(), offsetof(user, u_debugreg[6]), 0);
		state.dr[7] = ptrace(PTRACE_PEEKUSER, activeThread(), offsetof(user, u_debugreg[7]), 0);

	} else {
		state.clear();
	}
}

//------------------------------------------------------------------------------
// Name: setState(const State &state)
// Desc: 
//------------------------------------------------------------------------------
void DebuggerCore::setState(const State &state) {
	
	// TOOD: assert that we are paused
	
	if(attached()) {
		user_regs_struct regs;
#if defined(EDB_X86)
		regs.eax = state.eax;
		regs.ebx = state.ebx;
		regs.ecx = state.ecx;
		regs.edx = state.edx;
		regs.esp = state.esp;
		regs.ebp = state.ebp;
		regs.edi = state.edi;
		regs.esi = state.esi;
		regs.eip = state.eip;
		regs.eflags = state.eflags;
		regs.xcs = state.cs_;
		regs.xds = state.ds_;
		regs.xes = state.es_;
		regs.xfs = state.fs_;
		regs.xgs = state.gs_;
		regs.xss = state.ss_;
		regs.orig_eax = state.orig_eax;
#elif defined(EDB_X86_64)
		regs.rax = state.rax;
		regs.rbx = state.rbx;
		regs.rcx = state.rcx;
		regs.rdx = state.rdx;
		regs.rsp = state.rsp;
		regs.rbp = state.rbp;
		regs.rdi = state.rdi;
		regs.rsi = state.rsi;
		regs.r8 = state.r8;
		regs.r9 = state.r9;
		regs.r10 = state.r10;
		regs.r11 = state.r11;
		regs.r12 = state.r12;
		regs.r13 = state.r13;
		regs.r14 = state.r14;
		regs.r15 = state.r15;
		regs.rip = state.rip;
		regs.eflags = state.rflags;
		regs.cs = state.cs;
		regs.ds = state.ds;
		regs.es = state.es;
		regs.fs = state.fs;
		regs.gs = state.gs;
		regs.ss = state.ss;
		regs.orig_rax = state.orig_rax;
		regs.fs_base = state.fs_base;
		regs.gs_base = state.gs_base;
#endif
		ptrace(PTRACE_SETREGS, activeThread(), 0, &regs);

		ptrace(PTRACE_POKEUSER, activeThread(), offsetof(user, u_debugreg[0]), state.dr[0]);
		ptrace(PTRACE_POKEUSER, activeThread(), offsetof(user, u_debugreg[1]), state.dr[1]);
		ptrace(PTRACE_POKEUSER, activeThread(), offsetof(user, u_debugreg[2]), state.dr[2]);
		ptrace(PTRACE_POKEUSER, activeThread(), offsetof(user, u_debugreg[3]), state.dr[3]);
//		ptrace(PTRACE_POKEUSER, activeThread(), offsetof(user, u_debugreg[4]), state.dr[4]);
//		ptrace(PTRACE_POKEUSER, activeThread(), offsetof(user, u_debugreg[5]), state.dr[5]);
		ptrace(PTRACE_POKEUSER, activeThread(), offsetof(user, u_debugreg[6]), state.dr[6]);
		ptrace(PTRACE_POKEUSER, activeThread(), offsetof(user, u_debugreg[7]), state.dr[7]);
	}
}

//------------------------------------------------------------------------------
// Name: open(const QString &path, const QString &cwd, const QStringList &args, const QString &tty)
// Desc: 
//------------------------------------------------------------------------------
bool DebuggerCore::open(const QString &path, const QString &cwd, const QStringList &args, const QString &tty) {
	detach();
	pid_t pid;
	
	switch(pid = fork()) {
	case 0:
		// we are in the child now...
	
		// set ourselves (the child proc) up to be traced
		ptrace(PTRACE_TRACEME, 0, 0, 0);
		
		// redirect it's I/O
		if(!tty.isEmpty()) {
			FILE *const std_out	= freopen(qPrintable(tty), "r+b", stdout);
			FILE *const std_in	= freopen(qPrintable(tty), "r+b", stdin);
			FILE *const std_err	= freopen(qPrintable(tty), "r+b", stderr);
			
			Q_UNUSED(std_out);
			Q_UNUSED(std_in);
			Q_UNUSED(std_err);
		}
		
		// do the actual exec
		executeProcess(path, cwd, args);
		
		// we should never get here!
		abort();
		break;
	case -1:
		// error!
		m_PID = 0;
		return false;
	default:
		// parent
		do {
			m_Threads.clear();
			
			int status;
			if(native_waitpid(pid, &status, __WALL) == -1) {
				return false;
			}
			
			// the very first event should be a STOP of type SIGTRAP
			if(!WIFSTOPPED(status) || WSTOPSIG(status) != SIGTRAP) {
				detach();
				return false;
			}
			
#ifdef DEBUG_THREADS
			// enable following clones (threads)
			if(ptrace(PTRACE_SETOPTIONS, pid, 0, PTRACE_O_TRACECLONE) == -1) {
				detach();
				perror("PTRACE_SETOPTIONS");
				return false;
			}
#endif

			// setup the first event data for the primary thread
			m_Threads.insert(pid, ThreadInfo());
			m_PID					= pid;
			m_ActiveThread			= pid;
			m_Threads[pid].status	= status;
			return true;
		} while(0);
		break;
	}
}

void DebuggerCore::setActiveThread(edb::tid_t tid) {
	Q_ASSERT(m_Threads.contains(tid));
	m_ActiveThread = tid;

}

Q_EXPORT_PLUGIN2(DebuggerCore, DebuggerCore)
