/*
Copyright (C) 2006 - 2009 Evan Teran
                   eteran@alum.rit.edu

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "MD5.h"
#include <cstring>

//------------------------------------------------------------------------------
// Name: MD5()
// Desc: constructor
//------------------------------------------------------------------------------
MD5::MD5() : m_Count64(0), m_Finalized(false) {

	// Load magic initialization constants
	m_State[0] = 0x67452301;
	m_State[1] = 0xefcdab89;
	m_State[2] = 0x98badcfe;
	m_State[3] = 0x10325476;
}

//------------------------------------------------------------------------------
// Name: digest() const
// Desc: returns the final digest, only valid after being finalized
//------------------------------------------------------------------------------
const quint8 *MD5::digest() const {
	Q_ASSERT(m_Finalized);
	return reinterpret_cast<const quint8 *>(m_State);
}

//------------------------------------------------------------------------------
// Name: finalize()
// Desc: finalizes the state, making digest valid
//------------------------------------------------------------------------------
void MD5::finalize() {

	static const quint8 PADDING[64] = {
		0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
	};

	Q_ASSERT(!m_Finalized);
	
	const quint64 bits(m_Count64);
	
	// Pad out to 56 mod 64.
	quint32 padding = (56 - m_Count64 / CHAR_BIT) % sizeof(m_Buffer);
	if(padding == 0) {
		padding = sizeof(m_Buffer);
	}
	update(PADDING, padding);

	// append length (before padding)
	update(&bits, sizeof(bits));

	// clear out sensitive information
	std::memset(m_Buffer, 0, sizeof(m_Buffer));

	m_Finalized = true;
}

//------------------------------------------------------------------------------
// Name: update(const void *p, std::size_t input_length)
// Desc: MD5 block update operation. Continues an MD5 message-digest operation, 
//       processing another message block, and updating the context.
//------------------------------------------------------------------------------
void MD5::update(const void *p, std::size_t input_length) {

	Q_ASSERT(!m_Finalized);

	if(input_length != 0) {
		const quint8 *input = reinterpret_cast<const quint8 *>(p);

		std::size_t buffer_index;

		// Compute number of bytes mod 64
		buffer_index = (m_Count64 / CHAR_BIT) % sizeof(m_Buffer);

		m_Count64 += input_length * CHAR_BIT;

		// reduce buffering whenever we can, here's a shortcut
		if(buffer_index == 0 && (input_length % sizeof(m_Buffer)) == 0) {
			while(input_length != 0) {
				transform(input);
				input += sizeof(m_Buffer);
				input_length -= sizeof(m_Buffer);
			}

		} else {
		
			// how much space is left in buffer
			const quint32 buffer_space = sizeof(m_Buffer) - buffer_index;
			std::size_t input_index = 0;
			
			// do we have enough to fill the buffer?
			if (input_length >= buffer_space) {

				// fill the rest of the buffer and transform
				std::memcpy(&m_Buffer[buffer_index], input, buffer_space);
				transform(m_Buffer);

				// now, transform each 64-byte piece of the input, bypassing the buffer until we are out of blocks
				for(input_index = buffer_space; input_index + 63 < input_length; input_index += sizeof(m_Buffer)) {
					transform(input + input_index);
				}

				buffer_index = 0;	// so we can buffer remaining
			}

			if(input_length - input_index > 0) {
				// and here we do the buffering of the left over
				Q_ASSERT(input_length - input_index < sizeof(m_Buffer) - buffer_index);
				std::memcpy(&m_Buffer[buffer_index], &input[input_index], input_length - input_index);
			}
		}
	}
}

//------------------------------------------------------------------------------
// Name: transform(const quint8 block[64])
// Desc: MD5 basic transformation. Transforms state based on block.
//------------------------------------------------------------------------------
void MD5::transform(const quint8 block[64]) {

#define FF doTransform<F>
#define GG doTransform<G>
#define HH doTransform<H>
#define II doTransform<I>

	// Constants for MD5::transform routine
	static const int S11 = 7;
	static const int S12 = 12;
	static const int S13 = 17;
	static const int S14 = 22;
	static const int S21 = 5;
	static const int S22 = 9;
	static const int S23 = 14;
	static const int S24 = 20;
	static const int S31 = 4;
	static const int S32 = 11;
	static const int S33 = 16;
	static const int S34 = 23;
	static const int S41 = 6;
	static const int S42 = 10;
	static const int S43 = 15;
	static const int S44 = 21;

	quint32 a = m_State[0];
	quint32 b = m_State[1];
	quint32 c = m_State[2];
	quint32 d = m_State[3];

	const quint32 *const x = reinterpret_cast<const quint32 *>(block);

	Q_ASSERT(!m_Finalized);

	/* Round 1 */
	FF(a, b, c, d, x[ 0], S11, 0xd76aa478); /* 1 */
	FF(d, a, b, c, x[ 1], S12, 0xe8c7b756); /* 2 */
	FF(c, d, a, b, x[ 2], S13, 0x242070db); /* 3 */
	FF(b, c, d, a, x[ 3], S14, 0xc1bdceee); /* 4 */
	FF(a, b, c, d, x[ 4], S11, 0xf57c0faf); /* 5 */
	FF(d, a, b, c, x[ 5], S12, 0x4787c62a); /* 6 */
	FF(c, d, a, b, x[ 6], S13, 0xa8304613); /* 7 */
	FF(b, c, d, a, x[ 7], S14, 0xfd469501); /* 8 */
	FF(a, b, c, d, x[ 8], S11, 0x698098d8); /* 9 */
	FF(d, a, b, c, x[ 9], S12, 0x8b44f7af); /* 10 */
	FF(c, d, a, b, x[10], S13, 0xffff5bb1); /* 11 */
	FF(b, c, d, a, x[11], S14, 0x895cd7be); /* 12 */
	FF(a, b, c, d, x[12], S11, 0x6b901122); /* 13 */
	FF(d, a, b, c, x[13], S12, 0xfd987193); /* 14 */
	FF(c, d, a, b, x[14], S13, 0xa679438e); /* 15 */
	FF(b, c, d, a, x[15], S14, 0x49b40821); /* 16 */

	/* Round 2 */
	GG(a, b, c, d, x[ 1], S21, 0xf61e2562); /* 17 */
	GG(d, a, b, c, x[ 6], S22, 0xc040b340); /* 18 */
	GG(c, d, a, b, x[11], S23, 0x265e5a51); /* 19 */
	GG(b, c, d, a, x[ 0], S24, 0xe9b6c7aa); /* 20 */
	GG(a, b, c, d, x[ 5], S21, 0xd62f105d); /* 21 */
	GG(d, a, b, c, x[10], S22, 0x02441453); /* 22 */
	GG(c, d, a, b, x[15], S23, 0xd8a1e681); /* 23 */
	GG(b, c, d, a, x[ 4], S24, 0xe7d3fbc8); /* 24 */
	GG(a, b, c, d, x[ 9], S21, 0x21e1cde6); /* 25 */
	GG(d, a, b, c, x[14], S22, 0xc33707d6); /* 26 */
	GG(c, d, a, b, x[ 3], S23, 0xf4d50d87); /* 27 */
	GG(b, c, d, a, x[ 8], S24, 0x455a14ed); /* 28 */
	GG(a, b, c, d, x[13], S21, 0xa9e3e905); /* 29 */
	GG(d, a, b, c, x[ 2], S22, 0xfcefa3f8); /* 30 */
	GG(c, d, a, b, x[ 7], S23, 0x676f02d9); /* 31 */
	GG(b, c, d, a, x[12], S24, 0x8d2a4c8a); /* 32 */

	/* Round 3 */
	HH(a, b, c, d, x[ 5], S31, 0xfffa3942); /* 33 */
	HH(d, a, b, c, x[ 8], S32, 0x8771f681); /* 34 */
	HH(c, d, a, b, x[11], S33, 0x6d9d6122); /* 35 */
	HH(b, c, d, a, x[14], S34, 0xfde5380c); /* 36 */
	HH(a, b, c, d, x[ 1], S31, 0xa4beea44); /* 37 */
	HH(d, a, b, c, x[ 4], S32, 0x4bdecfa9); /* 38 */
	HH(c, d, a, b, x[ 7], S33, 0xf6bb4b60); /* 39 */
	HH(b, c, d, a, x[10], S34, 0xbebfbc70); /* 40 */
	HH(a, b, c, d, x[13], S31, 0x289b7ec6); /* 41 */
	HH(d, a, b, c, x[ 0], S32, 0xeaa127fa); /* 42 */
	HH(c, d, a, b, x[ 3], S33, 0xd4ef3085); /* 43 */
	HH(b, c, d, a, x[ 6], S34, 0x04881d05); /* 44 */
	HH(a, b, c, d, x[ 9], S31, 0xd9d4d039); /* 45 */
	HH(d, a, b, c, x[12], S32, 0xe6db99e5); /* 46 */
	HH(c, d, a, b, x[15], S33, 0x1fa27cf8); /* 47 */
	HH(b, c, d, a, x[ 2], S34, 0xc4ac5665); /* 48 */

	/* Round 4 */
	II(a, b, c, d, x[ 0], S41, 0xf4292244); /* 49 */
	II(d, a, b, c, x[ 7], S42, 0x432aff97); /* 50 */
	II(c, d, a, b, x[14], S43, 0xab9423a7); /* 51 */
	II(b, c, d, a, x[ 5], S44, 0xfc93a039); /* 52 */
	II(a, b, c, d, x[12], S41, 0x655b59c3); /* 53 */
	II(d, a, b, c, x[ 3], S42, 0x8f0ccc92); /* 54 */
	II(c, d, a, b, x[10], S43, 0xffeff47d); /* 55 */
	II(b, c, d, a, x[ 1], S44, 0x85845dd1); /* 56 */
	II(a, b, c, d, x[ 8], S41, 0x6fa87e4f); /* 57 */
	II(d, a, b, c, x[15], S42, 0xfe2ce6e0); /* 58 */
	II(c, d, a, b, x[ 6], S43, 0xa3014314); /* 59 */
	II(b, c, d, a, x[13], S44, 0x4e0811a1); /* 60 */
	II(a, b, c, d, x[ 4], S41, 0xf7537e82); /* 61 */
	II(d, a, b, c, x[11], S42, 0xbd3af235); /* 62 */
	II(c, d, a, b, x[ 2], S43, 0x2ad7d2bb); /* 63 */
	II(b, c, d, a, x[ 9], S44, 0xeb86d391); /* 64 */

	m_State[0] += a;
	m_State[1] += b;
	m_State[2] += c;
	m_State[3] += d;
	
#undef FF
#undef GG
#undef HH
#undef II
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
quint32 MD5::F(quint32 x, quint32 y, quint32 z) {
	return (x & y) | (~x & z);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
quint32 MD5::G(quint32 x, quint32 y, quint32 z) {
	return (z & x) | (~z & y);
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
quint32 MD5::H(quint32 x, quint32 y, quint32 z) {
	return x ^ y ^ z;
}

//------------------------------------------------------------------------------
// Name: 
// Desc: 
//------------------------------------------------------------------------------
quint32 MD5::I(quint32 x, quint32 y, quint32 z) {
	return y ^ (x | ~z);
}
